@@ -5,8 +5,15 @@
 """Entrypoint for the "toasty" command-line interface.
 
 """
+
 from __future__ import absolute_import, division, print_function
 
+__all__ = '''
+die
+entrypoint
+warn
+'''.split()
+
 import argparse
 import os.path
 import sys
@@ -31,6 +38,13 @@ def cascade_getparser(parser):
         type = int,
         help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
     )
+    parser.add_argument(
+        '--format', '-f',
+        metavar = 'FORMAT',
+        default = None,
+        choices = ['png', 'jpg', 'npy', 'fits'],
+        help = 'The format of data files to cascade. If not specified, this will be guessed.',
+    )
     parser.add_argument(
         '--start',
         metavar = 'DEPTH',
@@ -40,7 +54,7 @@ def cascade_getparser(parser):
     parser.add_argument(
         'pyramid_dir',
         metavar = 'DIR',
-        help = 'The directory containg the tile pyramid to cascade',
+        help = 'The directory containing the tile pyramid to cascade',
     )
 
 
@@ -49,7 +63,7 @@ def cascade_impl(settings):
     from .merge import averaging_merger, cascade_images
     from .pyramid import PyramidIO
 
-    pio = PyramidIO(settings.pyramid_dir)
+    pio = PyramidIO(settings.pyramid_dir, default_format=settings.format)
 
     start = settings.start
     if start is None:
@@ -57,7 +71,6 @@ def cascade_impl(settings):
 
     cascade_images(
         pio,
-        ImageMode.RGBA,
         start,
         averaging_merger,
         parallel=settings.parallelism,
@@ -65,46 +78,6 @@ def cascade_impl(settings):
     )
 
 
-# "healpix_sample_data_tiles" subcommand
-
-def healpix_sample_data_tiles_getparser(parser):
-    parser.add_argument(
-        '--outdir',
-        metavar = 'PATH',
-        default = '.',
-        help = 'The root directory of the output tile pyramid',
-    )
-    parser.add_argument(
-        'fitspath',
-        metavar = 'PATH',
-        help = 'The HEALPix FITS file to be tiled',
-    )
-    parser.add_argument(
-        'depth',
-        metavar = 'DEPTH',
-        type = int,
-        help = 'The depth of the TOAST layer to sample',
-    )
-
-
-def healpix_sample_data_tiles_impl(settings):
-    from .builder import Builder
-    from .image import ImageMode
-    from .pyramid import PyramidIO
-    from .samplers import healpix_fits_file_sampler
-
-    pio = PyramidIO(settings.outdir)
-    sampler = healpix_fits_file_sampler(settings.fitspath)
-    builder = Builder(pio)
-    builder.toast_base(ImageMode.F32, sampler, settings.depth)
-    builder.write_index_rel_wtml()
-
-    print(f'Successfully tiled input "{settings.fitspath}" at level {builder.imgset.tile_levels}.')
-    print('To create parent tiles, consider running:')
-    print()
-    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')
-
-
 # "make_thumbnail" subcommand
 
 def make_thumbnail_getparser(parser):
@@ -164,7 +137,7 @@ def multi_tan_make_data_tiles_impl(settings):
     from .multi_tan import MultiTanDataSource
     from .pyramid import PyramidIO
 
-    pio = PyramidIO(settings.outdir)
+    pio = PyramidIO(settings.outdir, default_format='npy')
     ds = MultiTanDataSource(settings.paths, hdu_index=settings.hdu_index)
     ds.compute_global_pixelization()
 
@@ -180,131 +153,9 @@ def multi_tan_make_data_tiles_impl(settings):
     # TODO: this should populate and emit a stub index_rel.wtml file.
 
 
-# "pipeline_fetch_inputs" subcommand
+# "pipeline" subcommands
 
-def _pipeline_add_io_args(parser):
-    parser.add_argument(
-        '--azure-conn-env',
-        metavar = 'ENV-VAR-NAME',
-        help = 'The name of an environment variable contain an Azure Storage '
-                'connection string'
-    )
-    parser.add_argument(
-        '--azure-container',
-        metavar = 'CONTAINER-NAME',
-        help = 'The name of a blob container in the Azure storage account'
-    )
-    parser.add_argument(
-        '--azure-path-prefix',
-        metavar = 'PATH-PREFIX',
-        help = 'A slash-separated path prefix for blob I/O within the container'
-    )
-    parser.add_argument(
-        '--local',
-        metavar = 'PATH',
-        help = 'Use the local-disk I/O backend'
-    )
-
-def _pipeline_io_from_settings(settings):
-    from .pipeline import AzureBlobPipelineIo, LocalPipelineIo
-
-    if settings.local:
-        return LocalPipelineIo(settings.local)
-
-    if settings.azure_conn_env:
-        conn_str = os.environ.get(settings.azure_conn_env)
-        if not conn_str:
-            die('--azure-conn-env=%s provided, but that environment variable is unset'
-                % settings.azure_conn_env)
-
-        if not settings.azure_container:
-            die('--azure-container-name must be provided if --azure-conn-env is')
-
-        path_prefix = settings.azure_path_prefix
-        if not path_prefix:
-            path_prefix = ''
-
-        return AzureBlobPipelineIo(
-            conn_str,
-            settings.azure_container,
-            path_prefix
-        )
-
-    die('An I/O backend must be specified with the arguments --local or --azure-*')
-
-
-def pipeline_fetch_inputs_getparser(parser):
-    _pipeline_add_io_args(parser)
-    parser.add_argument(
-        'workdir',
-        metavar = 'WORKDIR',
-        default = '.',
-        help = 'The local working directory',
-    )
-
-def pipeline_fetch_inputs_impl(settings):
-    from .pipeline import PipelineManager
-
-    pipeio = _pipeline_io_from_settings(settings)
-    mgr = PipelineManager(pipeio, settings.workdir)
-    mgr.fetch_inputs()
-
-
-# "pipeline_process_todos" subcommand
-
-def pipeline_process_todos_getparser(parser):
-    _pipeline_add_io_args(parser)
-    parser.add_argument(
-        'workdir',
-        metavar = 'WORKDIR',
-        default = '.',
-        help = 'The local working directory',
-    )
-
-def pipeline_process_todos_impl(settings):
-    from .pipeline import PipelineManager
-
-    pipeio = _pipeline_io_from_settings(settings)
-    mgr = PipelineManager(pipeio, settings.workdir)
-    mgr.process_todos()
-
-
-# "pipeline_publish_todos" subcommand
-
-def pipeline_publish_todos_getparser(parser):
-    _pipeline_add_io_args(parser)
-    parser.add_argument(
-        'workdir',
-        metavar = 'WORKDIR',
-        default = '.',
-        help = 'The local working directory',
-    )
-
-def pipeline_publish_todos_impl(settings):
-    from .pipeline import PipelineManager
-
-    pipeio = _pipeline_io_from_settings(settings)
-    mgr = PipelineManager(pipeio, settings.workdir)
-    mgr.publish_todos()
-
-
-# "pipeline_reindex" subcommand
-
-def pipeline_reindex_getparser(parser):
-    _pipeline_add_io_args(parser)
-    parser.add_argument(
-        'workdir',
-        metavar = 'WORKDIR',
-        default = '.',
-        help = 'The local working directory',
-    )
-
-def pipeline_reindex_impl(settings):
-    from .pipeline import PipelineManager
-
-    pipeio = _pipeline_io_from_settings(settings)
-    mgr = PipelineManager(pipeio, settings.workdir)
-    mgr.reindex()
+from .pipeline.cli import pipeline_getparser, pipeline_impl
 
 
 # "tile_allsky" subcommand
@@ -335,7 +186,7 @@ def tile_allsky_getparser(parser):
         metavar = 'PROJTYPE',
         default = 'plate-carree',
         help = 'The projection type of the input image (default: %(default)s; choices: %(choices)s)',
-        choices = ['plate-carree', 'plate-carree-galactic', 'plate-carree-planet'],
+        choices = ['plate-carree', 'plate-carree-galactic', 'plate-carree-ecliptic', 'plate-carree-planet'],
     )
     parser.add_argument(
         '--parallelism', '-j',
@@ -371,6 +222,9 @@ def tile_allsky_impl(settings):
     elif settings.projection == 'plate-carree-galactic':
         from .samplers import plate_carree_galactic_sampler
         sampler = plate_carree_galactic_sampler(img.asarray())
+    elif settings.projection == 'plate-carree-ecliptic':
+        from .samplers import plate_carree_ecliptic_sampler
+        sampler = plate_carree_ecliptic_sampler(img.asarray())
     elif settings.projection == 'plate-carree-planet':
         from .samplers import plate_carree_planet_sampler
         sampler = plate_carree_planet_sampler(img.asarray())
@@ -387,7 +241,6 @@ def tile_allsky_impl(settings):
         builder.make_thumbnail_from_other(img)
 
     builder.toast_base(
-        img.mode,
         sampler,
         settings.depth,
         is_planet=is_planet,
@@ -403,6 +256,46 @@ def tile_allsky_impl(settings):
     print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')
 
 
+# "tile_healpix" subcommand
+
+def tile_healpix_getparser(parser):
+    parser.add_argument(
+        '--outdir',
+        metavar = 'PATH',
+        default = '.',
+        help = 'The root directory of the output tile pyramid',
+    )
+    parser.add_argument(
+        'fitspath',
+        metavar = 'PATH',
+        help = 'The HEALPix FITS file to be tiled',
+    )
+    parser.add_argument(
+        'depth',
+        metavar = 'DEPTH',
+        type = int,
+        help = 'The depth of the TOAST layer to sample',
+    )
+
+
+def tile_healpix_impl(settings):
+    from .builder import Builder
+    from .image import ImageMode
+    from .pyramid import PyramidIO
+    from .samplers import healpix_fits_file_sampler
+
+    pio = PyramidIO(settings.outdir, default_format='npy')
+    sampler = healpix_fits_file_sampler(settings.fitspath)
+    builder = Builder(pio)
+    builder.toast_base(sampler, settings.depth)
+    builder.write_index_rel_wtml()
+
+    print(f'Successfully tiled input "{settings.fitspath}" at level {builder.imgset.tile_levels}.')
+    print('To create parent tiles, consider running:')
+    print()
+    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')
+
+
 # "tile_study" subcommand
 
 def tile_study_getparser(parser):
@@ -439,9 +332,13 @@ def tile_study_impl(settings):
     from .pyramid import PyramidIO
 
     img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
-    pio = PyramidIO(settings.outdir)
+    pio = PyramidIO(settings.outdir, default_format=img.default_format)
     builder = Builder(pio)
-    builder.default_tiled_study_astrometry()
+
+    if img.wcs is None:
+        builder.default_tiled_study_astrometry()
+    else:
+        builder.apply_wcs_info(img.wcs, img.width, img.height)
 
     # Do the thumbnail first since for large inputs it can be the memory high-water mark!
     if settings.placeholder_thumbnail:
@@ -507,6 +404,58 @@ def tile_wwtl_impl(settings):
     print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')
 
 
+# "transform" subcommand
+
+def transform_getparser(parser):
+    subparsers = parser.add_subparsers(dest='transform_command')
+
+    parser = subparsers.add_parser('fx3-to-rgb')
+    parser.add_argument(
+        '--parallelism', '-j',
+        metavar = 'COUNT',
+        type = int,
+        help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
+    )
+    parser.add_argument(
+        '--start',
+        metavar = 'DEPTH',
+        type = int,
+        help = 'The depth of the pyramid layer to start the cascade',
+    )
+    parser.add_argument(
+        '--clip', '-c',
+        metavar = 'NUMBER',
+        type = float,
+        default = 1.0,
+        help = 'The level at which to start flipping the floating-point data',
+    )
+    parser.add_argument(
+        'pyramid_dir',
+        metavar = 'DIR',
+        help = 'The directory containing the tile pyramid to cascade',
+    )
+
+
+def transform_impl(settings):
+    from .pyramid import PyramidIO
+
+    if settings.transform_command is None:
+        print('Run the "transform" command with `--help` for help on its subcommands')
+        return
+
+    if settings.transform_command == 'fx3-to-rgb':
+        from .transform import f16x3_to_rgb
+        pio = PyramidIO(settings.pyramid_dir)
+        f16x3_to_rgb(
+            pio, settings.start,
+            clip = settings.clip,
+            parallel = settings.parallelism,
+            cli_progress = True,
+        )
+    else:
+        die('unrecognized "transform" subcommand ' + settings.transform_command)
+
+
 # The CLI driver:
 
 def entrypoint(args=None):
