@@ -23,6 +23,44 @@ from PIL import Image as pil_image
 import numpy as np
 import sys
 
+try:
+    from astropy.io import fits
+    ASTROPY_INSTALLED = True
+except ImportError:
+    ASTROPY_INSTALLED = False
+
+PIL_RGB_FORMATS = {'jpg': 'JPEG'}
+PIL_RGBA_FORMATS = {'png': 'PNG'}
+PIL_FORMATS = PIL_RGB_FORMATS.copy()
+PIL_FORMATS.update(PIL_RGBA_FORMATS)
+SUPPORTED_FORMATS = list(PIL_RGB_FORMATS) + list(PIL_RGBA_FORMATS) + ['npy']
+
+if ASTROPY_INSTALLED:
+    SUPPORTED_FORMATS += ['fits']
+
+
+def _validate_format(name, fmt):
+    if fmt is not None and fmt not in SUPPORTED_FORMATS:
+        raise ValueError('{0} should be one of {1}'.format(name, '/'.join(sorted(SUPPORTED_FORMATS))))
+
+
+def _array_to_mode(array):
+    if array.ndim == 2 and array.dtype.kind == 'f':
+        if array.dtype.itemsize == 4:
+            return ImageMode.F32
+        elif array.dtype.itemsize == 8:
+            return ImageMode.F64
+    elif array.ndim == 3:
+        if array.shape[2] == 3:
+            if array.dtype.kind == 'f' and array.itemsize == 2:
+                return ImageMode.F16x3
+            elif array.dtype.kind == 'u' and array.dtype.itemsize == 1:
+                return ImageMode.RGB
+        elif array.shape[2] == 4:
+            if array.dtype.kind == 'u' and array.dtype.itemsize == 1:
+                return ImageMode.RGBA
+    raise ValueError('Could not determine mode for array with dtype {0} and shape {1}'.format(array.dtype, array.shape))
+
 
 class ImageMode(Enum):
     """
@@ -36,24 +74,8 @@ class ImageMode(Enum):
     RGB = 'RGB'
     RGBA = 'RGBA'
     F32 = 'F'
-
-    def get_default_save_extension(self):
-        """
-        Get the file extension to be used in this mode's "default" save format.
-
-        Returns
-        -------
-        The extension, without a period; either "png" or "npy"
-
-        """
-        # For RGB, could use JPG? But would complexify lots of things downstream.
-
-        if self in (ImageMode.RGB, ImageMode.RGBA):
-            return 'png'
-        elif self == ImageMode.F32:
-            return 'npy'
-        else:
-            raise Exception('unhandled mode in get_default_save_extension')
+    F64 = 'D'
+    F16x3 = 'F16x3'
 
     def make_maskable_buffer(self, buf_height, buf_width):
         """
@@ -79,17 +101,19 @@ class ImageMode(Enum):
         will be able to accept NaNs.
 
         """
-        mask_mode = self
 
         if self in (ImageMode.RGB, ImageMode.RGBA):
             arr = np.empty((buf_height, buf_width, 4), dtype=np.uint8)
-            mask_mode = ImageMode.RGBA
         elif self == ImageMode.F32:
             arr = np.empty((buf_height, buf_width), dtype=np.float32)
+        elif self == ImageMode.F64:
+            arr = np.empty((buf_height, buf_width), dtype=np.float64)
+        elif self == ImageMode.F16x3:
+            arr = np.empty((buf_height, buf_width, 3), dtype=np.float16)
         else:
             raise Exception('unhandled mode in make_maskable_buffer()')
 
-        return Image.from_array(mask_mode, arr)
+        return Image.from_array(arr)
 
     def try_as_pil(self):
         """
@@ -99,6 +123,8 @@ class ImageMode(Enum):
         -------
         A PIL image mode string, or None if there is no exact counterpart.
         """
+        if self == ImageMode.F16x3:
+            return None
         return self.value
 
 
@@ -108,13 +134,10 @@ class ImageLoader(object):
 
     This is implemented as its own class since there can be some options
     involved, and we want to provide a centralized place for handling them all.
-
-    TODO: support FITS, Numpy, etc.
-
     """
     black_to_transparent = False
     colorspace_processing = 'srgb'
-    desired_mode = None
+    crop = None
     psd_single_layer = None
 
     @classmethod
@@ -150,7 +173,11 @@ class ImageLoader(object):
             help = 'What kind of RGB colorspace processing to perform (default: %(default)s; choices: %(choices)s)',
             choices = ['srgb', 'none'],
         )
-        # not exposing desired_mode -- shouldn't be something the for the user to deal with
+        parser.add_argument(
+            '--crop',
+            metavar = 'TOP,RIGHT,BOTTOM,LEFT',
+            help = 'Crop the input image by discarding pixels from each edge (default: 0,0,0,0)',
+        )
         parser.add_argument(
             '--psd-single-layer',
             type = int,
@@ -177,6 +204,24 @@ class ImageLoader(object):
         loader.black_to_transparent = settings.black_to_transparent
         loader.colorspace_processing = settings.colorspace_processing
         loader.psd_single_layer = settings.psd_single_layer
+
+        if settings.crop is not None:
+            try:
+                crop = list(map(int, settings.crop.split(',')))
+                assert all(c >= 0 for c in crop)
+                assert len(crop) in (1, 2, 4)
+            except Exception:
+                raise Exception('cannot parse `--crop` setting `{settings.crop!r}`: should be a comma-separated list of 1, 2, or 4 non-negative integers')
+
+            if len(crop) == 1:
+                c = crop[0]
+                crop = [c, c, c, c]
+            elif len(crop) == 2:
+                cv, ch = crop
+                crop = [cv, ch, cv, ch]
+
+            loader.crop = crop
+
         return loader
 
     def load_pil(self, pil_img):
@@ -199,14 +244,35 @@ class ImageLoader(object):
         configuration.
 
         """
-        # If 8-bit grayscale, convert to RGB. Added for
-        # https://www.flickr.com/photos/10795027@N08/43023455582 . That file
-        # also comes with an ICC profile so we do the conversion before
-        # colorspace processing. TODO: if/when we do FITS to RGB conversion, we
-        # should probably use that codepath rather than this manual hack.
+        # If we're cropping, do it.
+        if self.crop is not None:
+            upper = self.crop[0]
+            right = pil_img.width - self.crop[1]
+            lower = pil_img.height - self.crop[2]
+            left = self.crop[3]
 
-        if pil_img.mode == 'L':
-            pil_img = pil_img.convert('RGBA')
+            # This operation can trigger PIL decompression-bomb aborts, so
+            # avoid them in the standard, non-thread-safe, way.
+            old_max = pil_image.MAX_IMAGE_PIXELS
+            try:
+                pil_image.MAX_IMAGE_PIXELS = None
+                pil_img = pil_img.crop((left, upper, right, lower))
+            finally:
+                pil_image.MAX_IMAGE_PIXELS = old_max
+
+        # If an unrecognized mode, try to standardize it. The weird PIL modes
+        # don't generally support an alpha channel, so we convert to RGB.
+
+        try:
+            ImageMode(pil_img.mode)
+        except ValueError:
+            print('warning: trying to convert image file to RGB from unexpected bitmode "%s"' % pil_img.mode)
+
+            if self.black_to_transparent:
+                # Avoid double-converting in the next filter.
+                pil_img = pil_img.convert('RGBA')
+            else:
+                pil_img = pil_img.convert('RGB')
 
         # Convert pure black to transparent -- make sure to do this before any
         # colorspace processing.
@@ -262,16 +328,6 @@ class ImageLoader(object):
                 xform = ImageCms.buildTransform(in_prof, out_prof, pil_img.mode, pil_img.mode)
                 ImageCms.applyTransform(pil_img, xform, inPlace=True)
 
-        # Make sure that we end up with the right mode, if requested.
-
-        if self.desired_mode is not None:
-            desired_pil = self.desired_mode.try_as_pil()
-            if desired_pil is None:
-                raise Exception('cannot convert PIL image to desired mode {}'.format(self.desired_mode))
-
-            if pil_img.mode != desired_pil:
-                pil_img = pil_img.convert(desired_pil)
-
         return Image.from_pil(pil_img)
 
     def load_stream(self, stream):
@@ -322,11 +378,29 @@ class ImageLoader(object):
 
         if path.endswith('.npy'):
             arr = np.load(path)
-            return Image.from_array(ImageMode.F32, arr.astype(np.float32))
+            return Image.from_array(arr, default_format='npy')
+
+        if path.lower().endswith(('.fits', '.fts', '.fits.gz', '.fts.gz')):
+
+            # TODO: implement a better way to recognize FITS files
+            # TODO: decide how to handle multiple HDUs
+            # TODO: decide how to handle non-TAN projections
+
+            hdu = fits.open(path)[0]
+            arr = hdu.data
+            if ASTROPY_INSTALLED:
+                from astropy.wcs import WCS
+                wcs = WCS(hdu.header)
+            else:
+                wcs = None
+
+            return Image.from_array(arr, wcs=wcs, default_format='fits')
 
         # Special handling for Photoshop files, used for some very large mosaics
         # with transparency (e.g. the PHAT M31/M33 images).
 
+        # TODO: check for AVM in following formats and set WCS using this if needed.
+
         if path.endswith('.psd') or path.endswith('.psb'):
             try:
                 from psd_tools import PSDImage
@@ -351,7 +425,9 @@ class ImageLoader(object):
         if path.endswith('.exr'):
             from .openexr import load_openexr
             img = load_openexr(path)
-            return Image.from_array(ImageMode.RGB, img)
+            if img.dtype != np.float16:
+                raise Exception('only half-precision OpenEXR images are currently supported')
+            return Image.from_array(img)
 
         # (One day, maybe we'll do more kinds of sniffing.) No special handling
         # came into play; just open the file and auto-detect.
@@ -367,12 +443,14 @@ class Image(object):
     have "bitmap" RGB(A) images and "science" floating-point images.
 
     """
-    _mode = None
     _pil = None
     _array = None
+    _mode = None
+    _default_format = 'png'
+    _wcs = None
 
     @classmethod
-    def from_pil(cls, pil_img):
+    def from_pil(cls, pil_img, wcs=None, default_format=None):
         """
         Create a new Image from a PIL image.
 
@@ -380,11 +458,19 @@ class Image(object):
         ----------
         pil_img : :class:`PIL.Image.Image`
             The source image.
+        wcs : :class:`~astropy.wcs.WCS`, optional
+            The WCS coordinate system for the image.
+        default_format : str, optional
+            The default format to use when writing the image if none is
+            specified explicitly.
 
         Returns
         -------
         A new :class:`Image` wrapping the PIL image.
         """
+
+        _validate_format('default_format', default_format)
+
         # Make sure that the image data are actually loaded from disk. Pillow
         # lazy-loads such that sometimes `np.asarray(img)` ends up failing
         # mysteriously if the image is loaded from a file handle that is closed
@@ -393,6 +479,8 @@ class Image(object):
 
         inst = cls()
         inst._pil = pil_img
+        inst._wcs = wcs
+        inst._default_format = default_format or cls._default_format
 
         try:
             inst._mode = ImageMode(pil_img.mode)
@@ -402,15 +490,21 @@ class Image(object):
         return inst
 
     @classmethod
-    def from_array(cls, mode, array):
+    def from_array(cls, array, wcs=None, default_format=None):
         """Create a new Image from an array-like data variable.
 
         Parameters
         ----------
-        mode : :class:`ImageMode`
-            The image mode.
         array : array-like object
-            The source data.
+            The source data. This should either be a 2-d floating point array a
+            3-d floating-point or uint8 array with shape (3, ny, nx), or a 3-d
+            uint8 array with shape (4, ny, nx).
+        wcs : :class:`~astropy.wcs.WCS`, optional
+            The WCS coordinate system for the image.
+        default_format : str, optional
+            The default format to use when writing the image if none is
+            specified explicitly. If not specified, this is automatically
+            chosen at write time based on the array type.
 
         Returns
         -------
@@ -418,29 +512,19 @@ class Image(object):
 
         Notes
         -----
-        The array will be converted to be at least two-dimensional. The data
-        array shape should match the requirements of the mode.
+        The array will be converted to be at least two-dimensional.
 
         """
+
+        _validate_format('default_format', default_format)
+
         array = np.atleast_2d(array)
-        array_ok = False
-
-        if mode == ImageMode.F32:
-            array_ok = (array.ndim == 2 and array.dtype == np.dtype(np.float32))
-        elif mode == ImageMode.RGB:
-            array_ok = (array.ndim == 3 and array.shape[2] == 3 and array.dtype == np.dtype(np.uint8))
-        elif mode == ImageMode.RGBA:
-            array_ok = (array.ndim == 3 and array.shape[2] == 4 and array.dtype == np.dtype(np.uint8))
-        else:
-            raise ValueError('unhandled image mode {} in from_array()'.format(mode))
-
-        if not array_ok:
-            raise ValueError('expected array compatible with mode {}, but got shape/dtype = {}/{}'
-                             .format(mode, array.shape, array.dtype))
 
         inst = cls()
-        inst._mode = mode
+        inst._mode = _array_to_mode(array)
+        inst._default_format = default_format or cls._default_format
         inst._array = array
+        inst._wcs = wcs
         return inst
 
     def asarray(self):
@@ -464,7 +548,6 @@ class Image(object):
             self._array = np.asarray(self._pil)
         return self._array
 
-
     def aspil(self):
         """Obtain the image data as :class:`PIL.Image.Image`.
 
@@ -479,6 +562,8 @@ class Image(object):
         """
         if self._pil is not None:
             return self._pil
+        if self.mode.try_as_pil() is None:
+            raise Exception(f'Toasty image with mode {self.mode} cannot be converted to PIL')
         return pil_image.fromarray(self._array)
 
     @property
@@ -490,6 +575,10 @@ class Image(object):
         # TODO: can this be more efficient? Does it need to be?
         return self.asarray().dtype
 
+    @property
+    def wcs(self):
+        return self._wcs
+
     @property
     def shape(self):
         if self._array is not None:
@@ -505,6 +594,23 @@ class Image(object):
     def height(self):
         return self.shape[0]
 
+    @property
+    def default_format(self):
+        if self._default_format is None:
+            if self.mode in (ImageMode.RGB, ImageMode.RGBA):
+                return 'png'
+            elif self.mode in (ImageMode.F32, ImageMode.F16x3):
+                return 'npy'
+        else:
+            return self._default_format
+
+    @default_format.setter
+    def default_format(self, value):
+        if value in SUPPORTED_FORMATS:
+            self._default_format = value
+        else:
+            raise ValueError('Unrecognized format: {0}'.format(value))
+
     def fill_into_maskable_buffer(self, buffer, iy_idx, ix_idx, by_idx, bx_idx):
         """
         Fill a maskable buffer with a rectangle of data from this image.
@@ -541,13 +647,60 @@ class Image(object):
         elif self.mode == ImageMode.RGBA:
             b.fill(0)
             b[by_idx,bx_idx] = i[iy_idx,ix_idx]
-        elif self.mode == ImageMode.F32:
+        elif self.mode in (ImageMode.F32, ImageMode.F16x3):
             b.fill(np.nan)
             b[by_idx,bx_idx] = i[iy_idx,ix_idx]
         else:
             raise Exception('unhandled mode in fill_into_maskable_buffer')
 
-    def save_default(self, path_or_stream):
+    def update_into_maskable_buffer(self, buffer, iy_idx, ix_idx, by_idx, bx_idx):
+        """
+        Update a maskable buffer with data from this image.
+
+        Parameters
+        ----------
+        buffer : :class:`Image`
+            The destination buffer image, created with :meth:`ImageMode.make_maskable_buffer`.
+        iy_idx : slice or other indexer
+            The indexer into the Y axis of the source image (self).
+        ix_idx : slice or other indexer
+            The indexer into the X axis of the source image (self).
+        by_idx : slice or other indexer
+            The indexer into the Y axis of the destination *buffer*.
+        bx_idx : slice or other indexer
+            The indexer into the X axis of the destination *buffer*.
+
+        Notes
+        -----
+        Unlike :meth:`fill_into_maskable_buffer`, this function does not clear
+        the entire buffer. It only overwrites the portion of the buffer covered
+        by non-NaN-like values of the input image.
+
+        """
+        i = self.asarray()
+        b = buffer.asarray()
+
+        sub_b = b[by_idx,bx_idx]
+        sub_i = i[iy_idx,ix_idx]
+
+        if self.mode == ImageMode.RGB:
+            sub_b[...,:3] = sub_i
+            sub_b[...,3] = 255
+        elif self.mode == ImageMode.RGBA:
+            valid = (sub_i[...,3] != 0)
+            valid = np.broadcast_to(valid[...,None], sub_i.shape)
+            np.putmask(sub_b, valid, sub_i)
+        elif self.mode == ImageMode.F32:
+            valid = ~np.isnan(sub_i)
+            np.putmask(sub_b, valid, sub_i)
+        elif self.mode == ImageMode.F16x3:
+            valid = ~np.any(np.isnan(sub_i), axis=2)
+            valid = np.broadcast_to(valid[...,None], sub_i.shape)
+            np.putmask(sub_b, valid, sub_i)
+        else:
+            raise Exception('unhandled mode in update_into_maskable_buffer')
+
+    def save(self, path_or_stream, format=None, mode=None):
         """
         Save this image to a filesystem path or stream
 
@@ -558,12 +711,25 @@ class Image(object):
             the stream should accept bytes.
 
         """
-        if self.mode in (ImageMode.RGB, ImageMode.RGBA):
-            self.aspil().save(path_or_stream, format='PNG')
-        elif self.mode == ImageMode.F32:
+
+        _validate_format('format', format)
+
+        format = format or self._default_format
+
+        if format in PIL_FORMATS:
+            pil_image = self.aspil()
+            if format in PIL_RGB_FORMATS and mode is None:
+                mode = ImageMode.RGB
+            if mode is not None:
+                pil_image = pil_image.convert(mode.try_as_pil())
+            pil_image.save(path_or_stream, format=PIL_FORMATS[format])
+        elif format == 'npy':
             np.save(path_or_stream, self.asarray())
-        else:
-            raise Exception('unhandled mode in save_default')
+        elif format == 'fits':
+            if self._wcs is None:
+                header = None if self._wcs is None else self._wcs.toheader()
+            fits.writeto(path_or_stream, self.asarray(),
+                         header=header, overwrite=True)
 
     def make_thumbnail_bitmap(self):
         """Create a thumbnail bitmap from the image.
@@ -575,7 +741,7 @@ class Image(object):
         be saved in JPEG format.
 
         """
-        if self.mode == ImageMode.F32:
+        if self.mode in (ImageMode.F32, ImageMode.F64, ImageMode.F16x3):
             raise Exception('cannot thumbnail-ify non-RGB Image')
 
         THUMB_SHAPE = (96, 45)
@@ -626,7 +792,7 @@ class Image(object):
         """
         if self._mode in (ImageMode.RGB, ImageMode.RGBA):
             self.asarray().fill(0)
-        elif self._mode == ImageMode.F32:
+        elif self._mode in (ImageMode.F32, ImageMode.F64, ImageMode.F16x3):
             self.asarray().fill(np.nan)
         else:
             raise Exception('unhandled mode in clear()')
